"""
@file
@brief Auto-regressor for timeseries.
"""
from .base import BaseTimeSeries, TimeSeriesRegressorMixin
from .dummies import DummyTimeSeriesRegressor


class ARTimeSeriesRegressor(BaseTimeSeries, TimeSeriesRegressorMixin):
    """
    Base class to build a regressor on timeseries.
    The class computes one or several predictions at each time,
    between *delay1* and *delay2*. It computes:
    :math:`\\hat{Y_{t+d} = f(Y_{t-1}, ..., Y_{t-p})}`
    with *d* in *[delay1, delay2[* and
    :math:`1 \\leqslant p \\leqslant past`.
    """

    def __init__(self, estimator="dummy", past=1, delay1=1, delay2=2,
                 use_all_past=False, preprocessing=None):
        """
        @param      estimator       estimator to use for regression,
                                    :epkg:`sklearn:linear_model:LinearRegression`
                                    implements a linear auto-regressor,
                                    ``'dummy'`` use past value as predictions
        @param      past            values to use to predict
        @param      delay1          the model computes the first prediction for
                                    *time=t + delay1*
        @param      delay2          the model computes the last prediction for
                                    *time=t + delay2* excluded
        @param      use_all_past    use all past features, not only the timeseries
        @param      preprocessing   preprocessing to apply before predicting,
                                    only the timeseries itselves, it can be
                                    a difference, it must be of type
                                    @see cl BaseReciprocalTimeSeriesTransformer
        """
        TimeSeriesRegressorMixin.__init__(self)
        BaseTimeSeries.__init__(self, past=past, delay1=delay1, delay2=delay2,
                                use_all_past=use_all_past, preprocessing=preprocessing)
        if estimator == "dummy":
            self.estimator = DummyTimeSeriesRegressor(
                past=past, delay1=delay1, delay2=delay2, use_all_past=use_all_past)
        if not hasattr(self.estimator, "fit"):
            raise TypeError(  # pragma: no cover
                "estimator is not an estimator but {}".format(type(estimator)))

    def fit(self, X, y, sample_weight=None):
        """
        Trains the model.

        :param X: output of
            X may be empty (None)
        :param y: timeseries (one single vector), array [n_obs]
        :param sample_weight: weights None or array [n_obs]
        :return: self
        """
        X, y, sample_weight = self._base_fit_predict(X, y, sample_weight)
        self.estimator_ = (self.estimator.fit(X, y)
                           if sample_weight is None
                           else self.estimator.fit(X, y, sample_weight=sample_weight))
        return self

    def predict(self, X, y):
        """
        Returns the prediction
        """
        X, y, _ = self._base_fit_predict(X, y, None)
        return self.estimator_.predict(X, y)



"""
@file
@brief Implements a custom criterion to train a decision tree.
"""
from libc.stdlib cimport calloc, free
from libc.stdio cimport printf
from libc.math cimport NAN

import numpy
cimport numpy
numpy.import_array()

from sklearn.tree._criterion cimport Criterion
from sklearn.tree._criterion cimport SIZE_t, DOUBLE_t


cdef class CommonRegressorCriterion(Criterion):
    """
    Common class to implement various version of `mean square error 
    <https://en.wikipedia.org/wiki/Mean_squared_error>`_.
    The code was inspired from
    `hellinger_distance_criterion.pyx 
    <https://github.com/EvgeniDubov/hellinger-distance-criterion/blob/master/
    hellinger_distance_criterion.pyx>`_,    
    `Cython example of exposing C-computed arrays in Python without data copies
    <http://gael-varoquaux.info/programming/
    cython-example-of-exposing-c-computed-arrays-in-python-without-data-copies.html>`_,
    `_criterion.pyx
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/tree/_criterion.pyx>`_.
    This implementation is not efficient but was made that way on purpose.
    It adds the features to the class.

    If the file does not compile, some explanations are given
    in :ref:`scikit-learn internal API
    <blog-internal-api-impurity-improvement>`_.
    """
    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    def __cinit__(self, const DOUBLE_t[:, ::1] X):
        self.sample_X = X

    cdef void _update_weights(self, SIZE_t start, SIZE_t end, SIZE_t old_pos, SIZE_t new_pos) nogil:
        """
        Unused.
        """
        pass

    cdef int reset(self) nogil except -1:
        """
        Resets the criterion at *pos=start*.
        This method must be implemented by the subclass.
        """
        self.pos = self.start

    cdef int reverse_reset(self) nogil except -1:
        """
        Resets the criterion at *pos=end*.
        This method must be implemented by the subclass.
        """
        self.pos = self.end

    cdef int update(self, SIZE_t new_pos) nogil except -1:
        """
        Updates statistics by moving ``samples[pos:new_pos]`` to the left child.
        This updates the collected statistics by moving ``samples[pos:new_pos]``
        from the right child to the left child. It must be implemented by
        the subclass.

        :param new_pos: SIZE_t
            New starting index position of the samples in the right child
        """
        self.pos = new_pos

    cdef void _mean(self, SIZE_t start, SIZE_t end, DOUBLE_t *mean, DOUBLE_t *weight) nogil:
        """
        Computes the mean of *y* between *start* and *end*.
        """
        raise NotImplementedError("Method _mean must be overloaded.")
            
    cdef double _mse(self, SIZE_t start, SIZE_t end, DOUBLE_t mean, DOUBLE_t weight) nogil:
        """
        Computes mean square error between *start* and *end*
        assuming corresponding points are approximated by a constant.
        """
        raise NotImplementedError("Method _mean must be overloaded.")
            
    cdef void children_impurity_weights(self, double* impurity_left,
                                        double* impurity_right,
                                        double* weight_left,
                                        double* weight_right) nogil:
        """
        Calculates the impurity of children,
        evaluates the impurity in
        children nodes, i.e. the impurity of ``samples[start:pos]``
        the impurity of ``samples[pos:end]``.

        :param impurity_left: double pointer
            The memory address where the impurity of the left child should be
            stored.
        :param impurity_right: double pointer
            The memory address where the impurity of the right child should be
            stored.
        :param weight_left: double pointer
            The memory address where the weight of the left child should be
            stored.
        :param weight_right: double pointer
            The memory address where the weight of the right child should be
            stored.
        """
        cdef DOUBLE_t mleft, mright
        self._mean(self.start, self.pos, &mleft, weight_left)
        self._mean(self.pos, self.end, &mright, weight_right)
        impurity_left[0] = self._mse(self.start, self.pos, mleft, weight_left[0])
        impurity_right[0] = self._mse(self.pos, self.end, mright, weight_right[0])

    ####################
    # functions used by a the tree optimizer
    ####################

    cdef double node_impurity(self) nogil:
        """
        Calculates the impurity of the node,
        the impurity of ``samples[start:end]``.
        This is the primary function of the criterion class.
        """
        cdef DOUBLE_t mean, weight
        self._mean(self.start, self.end, &mean, &weight)
        return self._mse(self.start, self.end, mean, weight)

    cdef void children_impurity(self, double* impurity_left,
                                double* impurity_right) nogil:
        """
        Calculates the impurity of children.
        
        :param impurity_left: double pointer
            The memory address where the impurity of the left child should be
            stored.
        :param impurity_right: double pointer
            The memory address where the impurity of the right child should be
            stored.
        """
        cdef DOUBLE_t wl, wr
        self.children_impurity_weights(impurity_left, impurity_right, &wl, &wr)

    cdef void node_value(self, double* dest) nogil:
        """
        Computes the node value, usually, the prediction
        the tree would do. Stores the value into *dest*.

        :param dest: double pointer
            The memory address where the node value should be stored.
        """
        cdef DOUBLE_t weight
        self._mean(self.start, self.end, dest, &weight)

    cdef double proxy_impurity_improvement(self) nogil:
        """
        Computes a proxy of the impurity reduction
        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.
        The absolute impurity improvement is only computed by the
        *impurity_improvement* method once the best split has been found.
        """
        cdef double impurity_left
        cdef double impurity_right
        self.children_impurity_weights(&impurity_left, &impurity_right,
                                       &self.weighted_n_left, &self.weighted_n_right)
        if self.pos == self.start or self.pos == self.end:
            return NAN

        return (- self.weighted_n_right * impurity_right
                - self.weighted_n_left * impurity_left)

    cdef double impurity_improvement(self, double impurity) nogil:
        """
        Computes the improvement in impurity
        This method computes the improvement in impurity when a split occurs.
        The weighted impurity improvement equation is the following::
        
            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where *N* is the total number of samples, *N_t* is the number of samples
        at the current node, *N_t_L* is the number of samples in the left child,
        and *N_t_R* is the number of samples in the right child,

        :param impurity: double
            The initial impurity of the node before the split
        :return: double, improvement in impurity after the split occurs
        """
        cdef double impurity_left
        cdef double impurity_right
        self.children_impurity_weights(&impurity_left, &impurity_right,
                                       &self.weighted_n_left, &self.weighted_n_right)
        if self.pos == self.start or self.pos == self.end:
            return NAN

        cdef double weight = self.weighted_n_left + self.weighted_n_right
        return ((weight / self.weighted_n_samples) *
                (impurity - (self.weighted_n_right / weight * impurity_right)
                          - (self.weighted_n_left / weight * impurity_left)))


def _test_criterion_init(Criterion criterion, 
                        const DOUBLE_t[:, ::1] y,
                        DOUBLE_t[:] sample_weight,
                        double weighted_n_samples,
                        SIZE_t[:] samples, 
                        SIZE_t start, SIZE_t end):
    "Test purposes. Methods cannot be directly called from python."
    criterion.init(y,
                   &sample_weight[0], weighted_n_samples,
                   &samples[0], start, end)


def _test_criterion_check(Criterion criterion):
    "Unused"
    pass


def assert_criterion_equal(Criterion c1, Criterion c2):
    "Unused"
    pass


def _test_criterion_node_impurity(Criterion criterion):
    "Test purposes. Methods cannot be directly called from python."
    return criterion.node_impurity()

    
def _test_criterion_proxy_impurity_improvement(Criterion criterion):
    "Test purposes. Methods cannot be directly called from python."
    return criterion.proxy_impurity_improvement()

    
def _test_criterion_impurity_improvement(Criterion criterion, double impurity):
    "Test purposes. Methods cannot be directly called from python."
    return criterion.impurity_improvement(impurity)

    
def _test_criterion_node_impurity_children(Criterion criterion):
    "Test purposes. Methods cannot be directly called from python."
    cdef DOUBLE_t left, right
    criterion.children_impurity(&left, &right)
    return left, right


def _test_criterion_node_value(Criterion criterion):
    "Test purposes. Methods cannot be directly called from python."
    cdef DOUBLE_t value
    criterion.node_value(&value)
    return value


def _test_criterion_update(Criterion criterion, SIZE_t new_pos):
    "Test purposes. Methods cannot be directly called from python."
    return criterion.update(new_pos)

    
def _test_criterion_printf(Criterion crit):
    "Test purposes. Methods cannot be directly called from python."    
    printf("start=%zu pos=%zu end=%zu\n", crit.start, crit.pos, crit.end)
    cdef DOUBLE_t left, right, value
    cdef int i;
    crit.children_impurity(&left, &right)
    crit.node_value(&value)
    printf("value: %f total=%f left=%f right=%f\n", value, 
           crit.node_impurity(), left, right)
    cdef int n = crit.y.shape[0]
    for i in range(0, n):
        printf("-- %d: y=%f\n", i, crit.y[i, 0])



# -*- coding: utf-8 -*-
"""
@file
@brief Implémente un *transform* qui suit la même API que tout :epkg:`scikit-learn` transform.
"""
import textwrap
import numpy
from .sklearn_base_transform import SkBaseTransform
from .sklearn_base_transform_learner import SkBaseTransformLearner


class SkBaseTransformStacking(SkBaseTransform):
    """
    Un *transform* qui cache plusieurs *learners*, arrangés
    selon la méthode du `stacking <http://blog.kaggle.com/2016/12/27/a-kagglers-guide-to-model-stacking-in-practice/>`_.

    .. exref::
        :title: Stacking de plusieurs learners dans un pipeline scikit-learn.
        :tag: sklearn
        :lid: ex-pipe2learner2

        Ce *transform* assemble les résultats de plusieurs learners.
        Ces features servent d'entrée à un modèle de stacking.

        .. runpython::
            :showcode:
            :warningout: FutureWarning

            from sklearn.model_selection import train_test_split
            from sklearn.datasets import load_iris
            from sklearn.linear_model import LogisticRegression
            from sklearn.tree import DecisionTreeClassifier
            from sklearn.metrics import accuracy_score
            from sklearn.pipeline import make_pipeline
            from mlinsights.sklapi import SkBaseTransformStacking

            data = load_iris()
            X, y = data.data, data.target
            X_train, X_test, y_train, y_test = train_test_split(X, y)

            trans = SkBaseTransformStacking([LogisticRegression(),
                                             DecisionTreeClassifier()])
            trans.fit(X_train, y_train)
            pred = trans.transform(X_test)
            print(pred[3:])
    """

    def __init__(self, models=None, method=None, **kwargs):
        """
        @param  models  list of learners
        @param  method  methods or list of methods to call
                        to convert features into prediction
                        (see below)
        @param  kwargs  parameters

        Available options for parameter *method*:

        * ``'predict'``
        * ``'predict_proba'``
        * ``'decision_function'``
        * a function

        If *method is None*, the default value is first
        ``predict_proba`` it it exists then ``predict``.
        """
        super().__init__(**kwargs)
        if models is None:
            raise ValueError("models cannot be None")  # pragma: no cover
        if not isinstance(models, list):
            raise TypeError(  # pragma: no cover
                "models must be a list not {0}".format(type(models)))
        if method is None:
            method = 'predict'
        if not isinstance(method, str):
            raise TypeError(  # pragma: no cover
                "Method must be a string not {0}".format(type(method)))
        self.method = method
        if isinstance(method, list):
            if len(method) != len(models):
                raise ValueError(  # pragma: no cover
                    "models and methods must have the same length: {0} != {1}".format(
                        len(models), len(method)))
        else:
            method = [method for m in models]

        def convert2transform(c, new_learners):
            "converting function into a transform"
            m, me = c
            if isinstance(m, SkBaseTransformLearner):
                if me == m.method:
                    return m
                res = SkBaseTransformLearner(m.model, me)
                new_learners.append(res)
                return res
            if hasattr(m, 'transform'):
                return m
            res = SkBaseTransformLearner(m, me)
            new_learners.append(res)
            return res

        new_learners = []
        res = list(map(lambda c: convert2transform(
            c, new_learners), zip(models, method)))
        if len(new_learners) == 0:
            # We need to do that to avoid creating new objects
            # when it is not necessary. This behavior is not
            # supported anymore by scikit-learn.
            # See sklearn.base.py
            self.models = models
        else:
            self.models = res

    def fit(self, X, y=None, **kwargs):
        """
        Trains a model.

        @param      X               features
        @param      y               targets
        @param      kwargs          additional parameters
        @return                     self
        """
        for m in self.models:
            m.fit(X, y=y, **kwargs)
        return self

    def transform(self, X):
        """
        Calls the learners predictions to convert
        the features.

        @param      X   features
        @return         prédictions
        """
        Xs = [m.transform(X) for m in self.models]
        return numpy.hstack(Xs)

    ##############
    # cloning API
    ##############

    def get_params(self, deep=True):
        """
        Returns the parameters which define the object.
        It follows :epkg:`scikit-learn` API.

        @param      deep        unused here
        @return                 dict
        """
        res = self.P.to_dict()
        res['models'] = self.models
        res['method'] = self.method
        if deep:
            for i, m in enumerate(self.models):
                par = m.get_params(deep)
                for k, v in par.items():
                    res["models_{0}__".format(i) + k] = v
        return res

    def set_params(self, **values):
        """
        Sets the parameters.

        @param      params      parameters
        """
        if 'models' in values:
            self.models = values['models']
            del values['models']
        if 'method' in values:
            self.method = values['method']
            del values['method']
        for k, v in values.items():
            if not k.startswith('models_'):
                raise ValueError(  # pragma: no cover
                    "Parameter '{0}' must start with 'models_'.".format(k))
        d = len('models_')
        pars = [{} for m in self.models]
        for k, v in values.items():
            si = k[d:].split('__', 1)
            i = int(si[0])
            pars[i][k[d + 1 + len(si):]] = v
        for p, m in zip(pars, self.models):
            if p:
                m.set_params(**p)

    #################
    # common methods
    #################

    def __repr__(self):
        """
        usual
        """
        rps = repr(self.P)
        res = "{0}([{1}], [{2}], {3})".format(
            self.__class__.__name__,
            ", ".join(repr(m.model if hasattr(m, 'model') else m)
                      for m in self.models),
            ", ".join(repr(m.method if hasattr(m, 'method') else None) for m in self.models), rps)
        return "\n".join(textwrap.wrap(res, subsequent_indent="    "))



"""
@file
@brief Implements a base class for a custom criterion to train a decision tree.
"""
cimport cython
import numpy
cimport numpy

numpy.import_array()

from libc.stdlib cimport calloc, free
from libc.math cimport NAN

from sklearn.tree._criterion cimport SIZE_t, DOUBLE_t
from ._piecewise_tree_regression_common cimport CommonRegressorCriterion


cdef class SimpleRegressorCriterion(CommonRegressorCriterion):
    """
    Implements `mean square error 
    <https://en.wikipedia.org/wiki/Mean_squared_error>`_
    criterion in a non efficient way. The code was inspired from
    `hellinger_distance_criterion.pyx 
    <https://github.com/EvgeniDubov/hellinger-distance-criterion/blob/master/
    hellinger_distance_criterion.pyx>`_,    
    `Cython example of exposing C-computed arrays in Python without data copies
    <http://gael-varoquaux.info/programming/
    cython-example-of-exposing-c-computed-arrays-in-python-without-data-copies.html>`_,
    `_criterion.pyx
    <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/tree/_criterion.pyx>`_.
    This implementation is not efficient but was made that way on purpose.
    It adds the features to the class.
    """
    cdef DOUBLE_t* sample_w
    cdef DOUBLE_t* sample_wy
    cdef SIZE_t* sample_i
    cdef DOUBLE_t sample_sum_wy
    cdef DOUBLE_t sample_sum_w

    def __dealloc__(self):
        """Destructor."""        
        free(self.sample_w)
        free(self.sample_wy)
        free(self.sample_i)
        self.sample_w = NULL
        self.sample_wy = NULL
        self.sample_i = NULL

    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    def __cinit__(self, const DOUBLE_t[:, ::1] X):
        self.sample_X = X
        # Allocate memory for the accumulators
        self.sample_w = NULL
        self.sample_wy = NULL
        self.sample_i = NULL
        
        # Criterion interface
        self.sample_weight = NULL
        self.samples = NULL
        self.sum_total = NULL
        self.sum_left = NULL
        self.sum_right = NULL

        # allocation
        if self.sample_w == NULL:
            self.sample_w = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_wy == NULL:
            self.sample_wy = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_i == NULL:
            self.sample_i = <SIZE_t*> calloc(X.shape[0], sizeof(SIZE_t))

    cdef int init(self, const DOUBLE_t[:, ::1] y,
                  DOUBLE_t* sample_weight,
                  double weighted_n_samples, SIZE_t* samples, 
                  SIZE_t start, SIZE_t end) nogil except -1:
        """
        This function is overwritten to check *y* and *X* size are the same.
        This API has changed in 0.21.
        """
        if y.shape[0] != self.sample_X.shape[0]:
            raise ValueError("X.shape={} -- y.shape={}".format(self.sample_X.shape, y.shape))
        if y.shape[1] != 1:
            raise ValueError("This class only works for a single vector.")
        return self.init_with_X(self.sample_X, y, sample_weight, weighted_n_samples,
                                samples, start, end)

    cdef int init_with_X(self, const DOUBLE_t[:, ::1] X, 
                         const DOUBLE_t[:, ::1] y,
                         DOUBLE_t* sample_weight,
                         double weighted_n_samples, SIZE_t* samples, 
                         SIZE_t start, SIZE_t end) nogil except -1:
        """
        Initializes the criterion.
        Returns -1 in case of failure to allocate memory
        (and raise *MemoryError*) or 0 otherwise.

        :param X: array-like, features, dtype=DOUBLE_t
        :param y: array-like, dtype=DOUBLE_t
            y is a buffer that can store values for n_outputs target variables
        :param sample_weight: array-like, dtype=DOUBLE_t
            The weight of each sample
        :param weighted_n_samples: DOUBLE_t
            The total weight of the samples being considered
        :param samples: array-like, dtype=DOUBLE_t
            Indices of the samples in X and y, where samples[start:end]
            correspond to the samples in this node
        :param start: SIZE_t
            The first sample to be used on this node
        :param end: SIZE_t
            The last sample used on this node
        """
        cdef SIZE_t ki, ks

        self.start = start
        self.pos = start
        self.end = end
        self.weighted_n_samples = weighted_n_samples
        self.y = y            

        self.sample_sum_wy = 0.
        self.sample_sum_w = 0.

        # Filling accumulators.
        for ki in range(<int>start, <int>end):
            ks = samples[ki]
            self.sample_i[ki] = ks
            self.sample_w[ki] = sample_weight[ks] if sample_weight else 1.
            self.sample_wy[ki] = self.sample_w[ki] * y[ks, 0]
            self.sample_sum_wy += y[ks, 0] * self.sample_w[ki]
            self.sample_sum_w += self.sample_w[ki]

        self.weighted_n_node_samples = self.sample_sum_w
        self.reset()
        if self.weighted_n_node_samples == 0:
            raise ValueError(
                "self.weighted_n_node_samples is null, first weight is %r." % self.sample_w[0])
        return 0

    cdef void _update_weights(self, SIZE_t start, SIZE_t end, SIZE_t old_pos, SIZE_t new_pos) nogil:
        """
        Updates members `weighted_n_right` and `weighted_n_left`
        when `pos` changes.
        """
        self.weighted_n_right = 0
        self.weighted_n_left = 0
        for k in range(start, new_pos):
            self.weighted_n_left += self.sample_w[k]
        for k in range(new_pos, end):
            self.weighted_n_right += self.sample_w[k]

    cdef void _mean(self, SIZE_t start, SIZE_t end, DOUBLE_t *mean, DOUBLE_t *weight) nogil:
        """
        Computes the mean of *y* between *start* and *end*.
        """
        if start == end:
            mean[0] = 0.
            return
        cdef DOUBLE_t m = 0.
        cdef DOUBLE_t w = 0.
        cdef int k
        for k in range(<int>start, <int>end):
            m += self.sample_wy[k]
            w += self.sample_w[k]
        weight[0] = w
        mean[0] = 0. if w == 0. else m / w
            
    cdef double _mse(self, SIZE_t start, SIZE_t end, DOUBLE_t mean, DOUBLE_t weight) nogil:
        """
        Computes mean square error between *start* and *end*
        assuming corresponding points are approximated by a constant.
        """
        if start == end:
            return 0.
        cdef DOUBLE_t squ = 0.
        cdef int k
        for k in range(<int>start, <int>end):            
            squ += (self.y[self.sample_i[k], 0] - mean) ** 2 * self.sample_w[k]
        return 0. if weight == 0. else squ / weight



"""
@file
@brief Datasets for timeseries.
"""
import datetime
import numpy
import pandas


def artificial_data(dt1, dt2, minutes=1):
    """
    Generates articial data every minutes.

    @param      dt1     first date
    @param      dt2     second date
    @param      minutes interval between two observations
    @return             dataframe

    .. runpython::
        :showcode:

        import datetime
        from mlinsights.timeseries.datasets import artificial_data

        now = datetime.datetime.now()
        data = artificial_data(now - datetime.timedelta(40), now)
        print(data.head())
    """

    def fxweek(x):
        return 2 - x * (1 - x)

    def sat(x):
        return 2 * x + 2

    data = []
    dt = datetime.timedelta(minutes=minutes)
    while dt1 < dt2:
        if dt1.weekday() == 6:
            dt1 += dt
            continue
        if minutes <= 120 and not (dt1.hour >= 8 and dt1.hour <= 18):
            dt1 += dt
            continue
        x = (dt1.hour - 8) / 10
        if dt1.weekday() == 5:
            y = sat(x)
        else:
            y = fxweek(x)
        data.append({'time': dt1, 'y': y})
        dt1 += dt
    df = pandas.DataFrame(data)
    df['y'] += numpy.random.randn(df.shape[0]) * 0.1
    df['time'] = pandas.DatetimeIndex(df['time'])
    return df



"""
@file
@brief Implements a custom criterion to train a decision tree.
"""
cimport cython
import numpy
cimport numpy

numpy.import_array()

from libc.stdlib cimport calloc, free
from libc.math cimport NAN

from sklearn.tree._criterion cimport SIZE_t, DOUBLE_t
from ._piecewise_tree_regression_common cimport CommonRegressorCriterion


cdef class SimpleRegressorCriterionFast(CommonRegressorCriterion):
    """
    Criterion which computes the mean square error
    assuming points falling into one node are approximated
    by a constant. The implementation follows the same
    design used in :class:`SimpleRegressorCriterion
    <mlinsights.mlmodel.piecewise_tree_regression_criterion.SimpleRegressorCriterion>`.
    This implementation is faster as it computes
    cumulated sums and avoids loops to compute
    intermediate gains.
    """
    cdef DOUBLE_t* sample_w_left
    cdef DOUBLE_t* sample_wy2_left
    cdef DOUBLE_t* sample_wy_left

    def __dealloc__(self):
        """Destructor."""        
        free(self.sample_w_left)
        free(self.sample_wy_left)
        free(self.sample_wy2_left)
        self.sample_w_left = NULL
        self.sample_wy_left = NULL
        self.sample_wy2_left = NULL

    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    def __cinit__(self, const DOUBLE_t[:, ::1] X):
        self.sample_X = X
        # Allocate memory for the accumulators
        self.sample_w_left = NULL
        self.sample_wy_left = NULL
        self.sample_wy2_left = NULL
        
        # Criterion interface
        self.sample_weight = NULL
        self.samples = NULL
        self.sum_total = NULL
        self.sum_left = NULL
        self.sum_right = NULL
        
        # allocations
        if self.sample_w_left == NULL:
            self.sample_w_left = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_wy_left == NULL:
            self.sample_wy_left = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_wy2_left == NULL:
            self.sample_wy2_left = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))

    cdef int init(self, const DOUBLE_t[:, ::1] y,
                  DOUBLE_t* sample_weight,
                  double weighted_n_samples, SIZE_t* samples, 
                  SIZE_t start, SIZE_t end) nogil except -1:
        """
        This function is overwritten to check *y* and *X* size are the same.
        This API has changed in 0.21.
        """
        if y.shape[0] != self.sample_X.shape[0]:
            raise ValueError("X.shape={} -- y.shape={}".format(self.sample_X.shape, y.shape))
        if y.shape[1] != 1:
            raise ValueError("This class only works for a single vector.")
        return self.init_with_X(self.sample_X, y, sample_weight, weighted_n_samples,
                                samples, start, end)

    cdef int init_with_X(self, const DOUBLE_t[:, ::1] X, 
                         const DOUBLE_t[:, ::1] y,
                         DOUBLE_t* sample_weight,
                         double weighted_n_samples, SIZE_t* samples, 
                         SIZE_t start, SIZE_t end) nogil except -1:
        """
        Initializes the criterion.
        Returns -1 in case of failure to allocate memory
        (and raise *MemoryError*) or 0 otherwise.

        :param X: array-like, features, dtype=DOUBLE_t
        :param y: array-like, dtype=DOUBLE_t
            y is a buffer that can store values for n_outputs target variables
        :param sample_weight: array-like, dtype=DOUBLE_t
            The weight of each sample
        :param weighted_n_samples: DOUBLE_t
            The total weight of the samples being considered
        :param samples: array-like, dtype=DOUBLE_t
            Indices of the samples in X and y, where samples[start:end]
            correspond to the samples in this node
        :param start: SIZE_t
            The first sample to be used on this node
        :param end: SIZE_t
            The last sample used on this node
        """
        cdef SIZE_t ki, ks
        cdef double w, y_

        self.start = start
        self.pos = start
        self.end = end
        self.weighted_n_samples = weighted_n_samples
        self.y = y            

        # we need to do that in case start > 0 or end < X.shape[0]
        for i in range(0, X.shape[0]):
            self.sample_w_left[i] = 0
            self.sample_wy_left[i] = 0
            self.sample_wy2_left[i] = 0

        # Left side.
        for ki in range(<int>start, <int>start+1):
            ks = samples[ki]
            w = sample_weight[ks] if sample_weight else 1.
            y_ = y[ks, 0]
            self.sample_w_left[ki] = w
            self.sample_wy_left[ki] = w * y_
            self.sample_wy2_left[ki] = w * y_ * y_
        for ki in range(<int>start+1, <int>end):
            ks = samples[ki]
            w = sample_weight[ks] if sample_weight else 1.
            y_ = y[ks, 0]
            self.sample_w_left[ki] = self.sample_w_left[ki-1] + w 
            self.sample_wy_left[ki] = self.sample_wy_left[ki-1] + w * y_
            self.sample_wy2_left[ki] = self.sample_wy2_left[ki-1] + w * y_ * y_
        
        self.weighted_n_node_samples = self.sample_w_left[end-1]
        self.reset()
        if self.weighted_n_node_samples == 0:
            raise ValueError(
                "self.weighted_n_node_samples is null, first weight is %r." % self.sample_w[0])
        return 0

    cdef void _mean(self, SIZE_t start, SIZE_t end, DOUBLE_t *mean, DOUBLE_t *weight) nogil:
        """
        Computes the mean of *y* between *start* and *end*.
        """
        if start == end:
            mean[0] = 0.
            return
        cdef DOUBLE_t m = self.sample_wy_left[end-1] - (self.sample_wy_left[start-1] if start > 0 else 0)
        cdef DOUBLE_t w = self.sample_w_left[end-1] - (self.sample_w_left[start-1] if start > 0 else 0)
        weight[0] = w
        mean[0] = 0. if w == 0. else m / w

    cdef double _mse(self, SIZE_t start, SIZE_t end, DOUBLE_t mean, DOUBLE_t weight) nogil:
        """
        Computes mean square error between *start* and *end*
        assuming corresponding points are approximated by a constant.
        """
        if start == end:
            return 0.
        cdef DOUBLE_t squ = self.sample_wy2_left[end-1] - (self.sample_wy2_left[start-1] if start > 0 else 0)
        # This formula only holds if mean is computed on the same interval.
        # Otherwise, it is squ / weight - true_mean ** 2 + (mean - true_mean) ** 2.
        return 0. if weight == 0. else squ / weight - mean ** 2

    cdef void _update_weights(self, SIZE_t start, SIZE_t end, SIZE_t old_pos, SIZE_t new_pos) nogil:
        """
        Updates members `weighted_n_right` and `weighted_n_left`
        when `pos` changes.
        """
        if new_pos == 0:
            self.weighted_n_left = 0.
            self.weighted_n_right = self.sample_w_left[end - 1]
        else:
            self.weighted_n_left = self.sample_w_left[new_pos - 1]
            self.weighted_n_right = self.sample_w_left[end - 1] - self.sample_w_left[new_pos - 1]



"""
@file
@brief Implements a way to get close examples based
on the output of a machine learned model.
"""
import numpy
from .search_engine_predictions import SearchEnginePredictions


class SearchEnginePredictionImages(SearchEnginePredictions):
    """
    Extends class @see cl SearchEnginePredictions.
    Vectors are coming from images. The metadata must contains
    information about path names. We assume all images can hold
    in memory. An example can found in notebook
    :ref:`searchimageskerasrst` or :ref:`searchimagestorchrst`.
    Another example can be found there:
    `search_images_dogcat.py
    <https://github.com/sdpython/ensae_projects/blob/master/src/
    ensae_projects/restapi/search_images_dogcat.py>`_.
    """

    def _prepare_fit(self, data=None, features=None, metadata=None,
                     transform=None, n=None, fLOG=None):
        """
        Stores data in the class itself.

        @param      data        a dataframe or None if the
                                the features and the metadata
                                are specified with an array and a
                                dictionary
        @param      features    features columns or an array
        @param      metadata    data
        @param      transform   transform each vector before using it
        @param      n           takes *n* images (or ``len(iter_images)``)
        @param      fLOG        logging function
        """
        if "torch" in str(type(data)):
            self.module_ = "torch"
            from torch.utils.data import DataLoader  # pylint: disable=E0401,C0415,E0611
            dataloader = DataLoader(
                data, batch_size=1, shuffle=False, num_workers=0)
            self.iter_images_ = iter_images = iter(
                zip(dataloader, data.samples))
            if n is None:
                n = len(data)
        elif "keras" in str(type(data)):
            self.module_ = "keras"
            iter_images = data
            # We delay the import as keras backend is not necessarily installed.
            from keras.preprocessing.image import Iterator  # pylint: disable=E0401,C0415,E0611
            from keras_preprocessing.image import DirectoryIterator, NumpyArrayIterator  # pylint: disable=E0401,C0415
            if not isinstance(iter_images, (Iterator, DirectoryIterator, NumpyArrayIterator)):
                raise NotImplementedError(  # pragma: no cover
                    "iter_images must be a keras Iterator. No option implemented for type {0}."
                    "".format(type(iter_images)))
            if iter_images.batch_size != 1:
                raise ValueError(  # pragma: no cover
                    "batch_size must be 1 not {0}".format(
                        iter_images.batch_size))
            self.iter_images_ = iter_images
            if n is None:
                n = len(iter_images)
            if not hasattr(iter_images, "filenames"):
                raise NotImplementedError(  # pragma: no cover
                    "Iterator does not iterate on images but numpy arrays (not implemented).")
        else:
            raise TypeError(  # pragma: no cover
                "Unexpected data type {0}.".format(type(data)))

        def get_current_index(flow):
            "get current index"
            return flow.index_array[(flow.batch_index + flow.n - 1) % flow.n]

        def iterator_feature_meta():
            "iterators on metadata"
            def accessor(iter_images):
                if hasattr(iter_images, 'filenames'):
                    # keras
                    return (lambda i, ite: (ite, iter_images.filenames[get_current_index(iter_images)]))
                else:
                    # torch
                    return (lambda i, ite: (ite[0], ite[1][0]))
            acc = accessor(iter_images)

            for i, it in zip(range(n), iter_images):
                im, name = acc(i, it)
                if not isinstance(name, str):
                    raise TypeError(  # pragma: no cover
                        "name should be a string, not {0}".format(type(name)))
                yield im[0], dict(name=name, i=i)
                if fLOG and i % 10000 == 0:
                    fLOG(
                        '[SearchEnginePredictionImages.fit] i={}/{} - {}'.format(i, n, name))
        super()._prepare_fit(data=iterator_feature_meta(), transform=transform)

    def fit(self, iter_images, n=None, fLOG=None):
        """
        Processes images through the model and fits a *k-nn*.

        @param      iter_images `Iterator <https://github.com/fchollet/keras/blob/master/keras/preprocessing/image.py#L719>`_
        @param      n           takes *n* images (or ``len(iter_images)``)
        @param      fLOG        logging function
        @param      kwimg       parameters used to preprocess the images
        """
        self._prepare_fit(data=iter_images, transform=self.fct, n=n, fLOG=fLOG)
        return self._fit_knn()

    def kneighbors(self, iter_images, n_neighbors=None):
        """
        Searches for neighbors close to the first image
        returned by *iter_images*. It returns the neighbors
        only for the first image.

        @param      iter_images `Iterator <https://github.com/fchollet/keras/blob/master/keras/preprocessing/image.py#L719>`_
        @return                 score, ind, meta

        *score* is an array representing the lengths to points,
        *ind* contains the indices of the nearest points in the population matrix,
        *meta* is the metadata.
        """
        if isinstance(iter_images, numpy.ndarray):
            if self.module_ == "keras":
                raise NotImplementedError("Not yet implemented or Keras.")
            elif self.module_ == "torch":
                from torch import from_numpy  # pylint: disable=E0611,E0401,C0415
                X = from_numpy(iter_images[numpy.newaxis, :, :, :])
                return super().kneighbors(X, n_neighbors=n_neighbors)
            raise RuntimeError(
                "Unknown module '{0}'.".format(self.module_))
        elif "keras" in str(iter_images):
            if self.module_ != "keras":
                raise RuntimeError(  # pragma: no cover
                    "Keras object but {0} was used to train the KNN.".format(self.module_))
            # We delay the import as keras backend is not necessarily installed.
            # keras, it expects an iterator.
            from keras.preprocessing.image import Iterator  # pylint: disable=E0401,C0415,E0611
            from keras_preprocessing.image import DirectoryIterator, NumpyArrayIterator  # pylint: disable=E0401,C0415,E0611
            if not isinstance(iter_images, (Iterator, DirectoryIterator, NumpyArrayIterator)):
                raise NotImplementedError(  # pragma: no cover
                    "iter_images must be a keras Iterator. No option implemented for type {0}.".format(type(iter_images)))
            if iter_images.batch_size != 1:
                raise ValueError(  # pragma: no cover
                    "batch_size must be 1 not {0}".format(
                        iter_images.batch_size))
            for img in iter_images:
                X = img[0]
                break
            return super().kneighbors(X, n_neighbors=n_neighbors)
        elif "torch" in str(type(iter_images)):
            if self.module_ != "torch":
                raise RuntimeError(  # pragma: no cover
                    "Torch object but {0} was used to train the KNN.".format(self.module_))
            # torch: it expects a tensor
            X = iter_images
            return super().kneighbors(X, n_neighbors=n_neighbors)
        elif isinstance(iter_images, list):
            res = [self.kneighbors(it, n_neighbors=n_neighbors)
                   for it in iter_images]
            return (numpy.vstack([_[0] for _ in res]),
                    numpy.vstack([_[1] for _ in res]),
                    numpy.vstack([_[2] for _ in res]))
        else:
            raise TypeError(  # pragma: no cover
                "Unexpected type {0} in SearchEnginePredictionImages.kneighbors".format(
                    type(iter_images)))



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a *learner* which follows the same API
as every :epkg:`scikit-learn` learner.
"""
from .sklearn_base import SkBase


class SkBaseLearner(SkBase):

    """
    Pattern of a *learner* qui suit la même API que :epkg:`scikit-learn`.
    """

    def __init__(self, **kwargs):
        """
        constructor
        """
        SkBase.__init__(self, **kwargs)

    ###################
    # API scikit-learn
    ###################

    def fit(self, X, y=None, sample_weight=None):
        """
        Trains a model.

        @param      X               features
        @param      y               targets
        @param      sample_weight   weight
        @return                     self
        """
        raise NotImplementedError()  # pragma: no cover

    def predict(self, X):
        """
        Predicts.

        @param      X   features
        @return         prédictions
        """
        raise NotImplementedError()  # pragma: no cover

    def decision_function(self, X):
        """
        Output of the model in case of a regressor,
        matrix with a score for each class and each sample
        for a classifier.

        @param      X   Samples, {array-like, sparse matrix}, shape = (n_samples, n_features)
        @return         array, shape = (n_samples,.), Returns predicted values.
        """
        raise NotImplementedError()  # pragma: no cover

    def score(self, X, y=None, sample_weight=None):
        """
        Returns the mean accuracy on the given test data and labels.

        @param      X               Training data, numpy array or sparse matrix of shape [n_samples,n_features]
        @param      y               Target values, numpy array of shape [n_samples, n_targets] (optional)
        @param      sample_weight   Weight values, numpy array of shape [n_samples, n_targets] (optional)
        @return                     score : float, Mean accuracy of self.predict(X) wrt. y.
        """
        raise NotImplementedError()  # pragma: no cover



"""
@file
@brief Implements new features such as polynomial features.
"""
from itertools import combinations, chain
from itertools import combinations_with_replacement as combinations_w_r


def _transform_iall(degree, bias, XP, X, multiply, final):
    "Computes the polynomial features"
    if bias:
        XP[:, 0] = 1
        pos = 1
    else:
        pos = 0

    n = X.shape[1]
    for d in range(0, degree):
        if d == 0:
            XP[:, pos:pos + n] = X
            index = list(range(pos, pos + n))
            pos += n
            index.append(pos)
        else:
            new_index = []
            end = index[-1]
            for i in range(0, n):
                a = index[i]
                new_index.append(pos)
                new_pos = pos + end - a
                multiply(XP[:, a:end], X[:, i:i + 1],
                         XP[:, pos:new_pos])
                pos = new_pos

            new_index.append(pos)
            index = new_index

    return final(XP)


def _transform_ionly(degree, bias, XP, X, multiply, final):
    "Computes the polynomial features"
    if bias:
        XP[:, 0] = 1
        pos = 1
    else:
        pos = 0

    n = X.shape[1]
    for d in range(0, degree):
        if d == 0:
            XP[:, pos:pos + n] = X
            index = list(range(pos, pos + n))
            pos += n
            index.append(pos)
        else:
            new_index = []
            end = index[-1]
            for i in range(0, n):
                a = index[i]
                new_index.append(pos)
                dec = index[i + 1] - index[i]
                new_pos = pos + end - a - dec
                if new_pos <= pos:
                    break
                multiply(XP[:, a + dec:end], X[:, i:i + 1],
                         XP[:, pos:new_pos])
                pos = new_pos

            new_index.append(pos)
            index = new_index

    return final(XP)


def _combinations_poly(n_features, degree, interaction_only, include_bias):
    "Computes all polynomial features combinations."
    comb = (combinations if interaction_only else combinations_w_r)
    start = int(not include_bias)
    return chain.from_iterable(comb(range(n_features), i)
                               for i in range(start, degree + 1))



"""
@file
@brief Correlations.
"""
import numpy
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import scale
from sklearn import clone


def non_linear_correlations(df, model, draws=5, minmax=False):
    """
    Computes non linear correlations.

    @param      df      :epkg:`pandas:DataFrame` or
                        :epkg:`numpy:array`
    @param      model   machine learned model used to compute
                        the correlations
    @param      draws   number of tries for :epkg:`bootstrap`,
                        the correlation is the average of the results
                        obtained at each draw
    @param      minmax  if True, returns three matrices correlations, min, max,
                        only the correlation matrix if False
    @return             see parameter minmax

    `Pearson Correlations <https://fr.wikipedia.org/wiki/Corr%C3%A9lation_(statistiques)>`_
    is:

    .. math::

        cor(X_i, X_j) = \\frac{cov(X_i, Y_i)}{\\sigma(X_i)\\sigma(X_j)}

    If variables are centered, :math:`\\mathbb{E}X_i=\\mathbb{E}X_j=0`,
    it becomes:

    .. math::

        cor(X_i, X_j) = \\frac{\\mathbb{E}(X_i X_j)}{\\sqrt{\\mathbb{E}X_i^2 \\mathbb{E}X_j^2}}

    If rescaled, :math:`\\mathbb{E}X_i^2=\\mathbb{E}X_j^2=1`,
    then it becomes :math:`cor(X_i, X_j) = \\mathbb{E}(X_i X_j)`.
    Let's assume we try to find a coefficient such as
    :math:`\\alpha_{ij}` minimizes the standard deviation
    of noise :math:`\\epsilon_{ij}`:

    .. math::

        X_j = \\alpha_{ij}X_i + \\epsilon_{ij}

    It is like if coefficient :math:`\\alpha_{ij}` comes from a
    a linear regression which minimizes
    :math:`\\mathbb{E}(X_j - \\alpha_{ij}X_i)^2`.
    If variable :math:`X_i`, :math:`X_j` are centered
    and rescaled: :math:`\\alpha_{ij}^* = \\mathbb{E}(X_i X_j) = cor(X_i, X_j)`.
    We extend that definition to function :math:`f` of parameter :math:`\\omega`
    defined as: :math:`f(\\omega, X) \\rightarrow \\mathbb{R}`.
    :math:`f` is not linear anymore.
    Let's assume parameter :math:`\\omega^*` minimizes
    quantity :math:`\\min_\\omega (X_j  - f(\\omega, X_i))^2`.
    Then :math:`X_j = \\alpha_{ij} \\frac{f(\\omega^*, X_i)}{\\alpha_{ij}} + \\epsilon_{ij}`
    and we choose :math:`\\alpha_{ij}` such as
    :math:`\\mathbb{E}\\left(\\frac{f(\\omega^*, X_i)^2}{\\alpha_{ij}^2}\\right) = 1`.
    Let's define a non linear correlation bounded by :math:`f` as:

    .. math::

        cor^f(X_i, X_j) = \\sqrt{ \\mathbb{E} (f(\\omega, X_i)^2 )}

    We can verify that this value is in interval`:math:`[0,1]``.
    That also means that there is no negative correlation.
    :math:`f` is a machine learned model and most of them
    usually overfit the data. The database is split into
    two parts, one is used to train the model, the other
    one to compute the correlation. The same split are used
    for every coefficient. The returned matrix is not
    necessarily symmetric.

    .. exref::
        :title: Compute non linear correlations

        The following example compute non linear correlations
        on :epkg:`Iris` datasets based on a
        :epkg:`RandomForestRegressor` model.

        .. runpython::
            :showcode:
            :warningout: FutureWarning

            import pandas
            from sklearn import datasets
            from sklearn.ensemble import RandomForestRegressor
            from mlinsights.metrics import non_linear_correlations

            iris = datasets.load_iris()
            X = iris.data[:, :4]
            df = pandas.DataFrame(X)
            df.columns = ["X1", "X2", "X3", "X4"]
            cor = non_linear_correlations(df, RandomForestRegressor())
            print(cor)

    """

    if hasattr(df, 'iloc'):
        cor = df.corr()
        cor.iloc[:, :] = 0.
        iloc = True
        if minmax:
            mini = cor.copy()
            maxi = cor.copy()
    else:
        cor = numpy.corrcoef(df, rowvar=False)
        cor[:, :] = 0.
        iloc = False
        if minmax:
            mini = cor.copy()
            maxi = cor.copy()
    df = scale(df)

    for k in range(0, draws):
        df_train, df_test = train_test_split(df, test_size=0.5)
        for i in range(cor.shape[0]):
            xi_train = df_train[:, i:i + 1]
            xi_test = df_test[:, i:i + 1]
            for j in range(cor.shape[1]):
                xj_train = df_train[:, j:j + 1]
                xj_test = df_test[:, j:j + 1]
                if len(xj_test) == 0 or len(xi_test) == 0:
                    raise ValueError(  # pragma: no cover
                        "One column is empty i={0} j={1}.".format(i, j))
                mod = clone(model)
                try:
                    mod.fit(xi_train, xj_train.ravel())
                except Exception as e:  # pragma: no cover
                    raise ValueError(
                        "Unable to compute correlation for i={0} j={1}.".format(i, j)) from e
                v = mod.predict(xi_test)
                c = (1 - numpy.var(v - xj_test.ravel()))
                co = max(c, 0) ** 0.5
                if iloc:
                    cor.iloc[i, j] += co
                    if minmax:
                        if k == 0:
                            mini.iloc[i, j] = co
                            maxi.iloc[i, j] = co
                        else:
                            mini.iloc[i, j] = min(mini.iloc[i, j], co)
                            maxi.iloc[i, j] = max(maxi.iloc[i, j], co)
                else:
                    cor[i, j] += co
                    if minmax:
                        if k == 0:
                            mini[i, j] = co
                            maxi[i, j] = co
                        else:
                            mini[i, j] = min(mini[i, j], co)
                            maxi[i, j] = max(maxi[i, j], co)
    if minmax:
        return cor / draws, mini, maxi
    return cor / draws



# -*- coding: utf-8 -*-
"""
@file
@brief Implements class @see cl SkBaseClassifier.
"""
from sklearn.metrics import accuracy_score
from .sklearn_base_learner import SkBaseLearner


class SkBaseClassifier(SkBaseLearner):
    """
    Defines a custom classifier.
    """

    def __init__(self, **kwargs):
        """
        constructor
        """
        SkBaseLearner.__init__(self, **kwargs)

    def score(self, X, y=None, sample_weight=None):
        """
        Returns the mean accuracy on the given test data and labels.

        @param      X               Training data, numpy array or sparse matrix of shape [n_samples,n_features]
        @param      y               Target values, numpy array of shape [n_samples, n_targets] (optional)
        @param      sample_weight   Weight values, numpy array of shape [n_samples, n_targets] (optional)
        @return                     score : float, Mean accuracy of self.predict(X) wrt. y.
        """
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)

    def predict_proba(self, X):
        """
        Returns probability estimates for the test data X.

        @param      X       Training data, numpy array or sparse matrix of shape [n_samples,n_features]
        @return             array, shape = (n_samples,.), Returns predicted values.
        """
        raise NotImplementedError()



"""
@file
@brief Implements a way to get close examples based
on the output of a machine learned model.
"""
from ..mlmodel import model_featurizer
from ..helpers.parameters import format_function_call
from .search_engine_vectors import SearchEngineVectors


class SearchEnginePredictions(SearchEngineVectors):
    """
    Extends class @see cl SearchEngineVectors by
    looking for neighbors to a vector *X* by
    looking neighbors to *f(X)* and not *X*.
    *f* can be any function which converts a vector
    into another one or a machine learned model.
    In that case, *f* will be set to a default behavior.
    See function @see fn model_featurizer.
    """

    def __init__(self, fct, fct_params=None, **knn):
        """
        @param      fct         function *f* applied before looking for neighbors,
                                it can also be a machine learned model
        @param      fct_params  parameters sent to function @see fn model_featurizer
        @param      pknn        list of parameters, see
                                :epkg:`sklearn:neighborsNearestNeighbors`
        """
        super().__init__(**knn)
        self._fct_params = fct_params
        self._fct_init = fct
        if (callable(fct) and not hasattr(fct, 'predict') and
                not hasattr(fct, 'forward')):
            self.fct = fct
        else:
            if fct_params is None:
                fct_params = {}
            self.fct = model_featurizer(fct, **fct_params)

    def __repr__(self):
        """
        usual
        """
        if self.pknn:
            pp = self.pknn.copy()
        else:
            pp = {}
        pp['fct'] = self._fct_init
        pp['fct_params'] = self._fct_params
        return format_function_call(self.__class__.__name__, pp)

    def fit(self, data=None, features=None, metadata=None):
        """
        Every vector comes with a list of metadata.

        @param      data        a :epkg:`dataframe` or None if the
                                the features and the metadata
                                are specified with an array and a
                                dictionary
        @param      features    features columns or an array
        @param      metadata    data
        """
        iterate = self._is_iterable(data)
        if iterate:
            self._prepare_fit(data=data, features=features,
                              metadata=metadata, transform=self.fct)
        else:
            self._prepare_fit(data=data, features=features, metadata=metadata)
            if isinstance(self.features_, list):
                raise TypeError(  # pragma: no cover
                    "features_ cannot be a list when training the model.")
            self.features_ = self.fct(self.features_, True)
        return self._fit_knn()

    def kneighbors(self, X, n_neighbors=None):
        """
        Searches for neighbors close to *X*.

        @param      X               features
        @return                     score, ind, meta

        *score* is an array representing the lengths to points,
        *ind* contains the indices of the nearest points in the population matrix,
        *meta* is the metadata.
        """
        xp = self.fct(X, False)
        if len(xp.shape) == 1:
            xp = xp.reshape((1, len(xp)))
        return super().kneighbors(xp, n_neighbors=n_neighbors)



"""
@file
@brief Timeseries data manipulations.
"""
import numpy
from sklearn import get_config


def build_ts_X_y(model, X, y, weights=None, same_rows=False):
    """
    Builds standard *X, y* based in the given one.

    @param      model       a timeseries model (@see cl BaseTimeSeries)
    @param      X           times series, used as features, [n_obs, n_features],
                            X may be empty (None)
    @param      y           timeseries (one single vector),  [n_obs]
    @param      weights     weights None or array [n_obs]
    @param      same_rows   keep the same number of rows
                            as the original datasets, use nan when no value is
                            available
    @return                 *(X, y, weights)*:  X is array of features [nrows, n_features + past]
                            where `nrows = n_obs + model.delay2 - model.past + 2`,
                            y is an array of targets [nrows],
                            weights is None or array [nrows]

    .. runpython::
        :showcode:

        import numpy
        from mlinsights.timeseries import build_ts_X_y
        from mlinsights.timeseries.base import BaseTimeSeries

        X = numpy.arange(10).reshape(5, 2)
        y = numpy.arange(5) * 100
        weights = numpy.arange(5) * 1000
        bs = BaseTimeSeries(past=2)
        nx, ny, nw = build_ts_X_y(bs, X, y, weights)
        print('X=', X)
        print('y=', y)
        print('nx=', nx)
        print('ny=', ny)

    With ``use_all_past=True``:

    .. runpython::
        :showcode:

        import numpy
        from mlinsights.timeseries.base import BaseTimeSeries
        from mlinsights.timeseries import build_ts_X_y

        X = numpy.arange(10).reshape(5, 2)
        y = numpy.arange(5) * 100
        weights = numpy.arange(5) * 1000
        bs = BaseTimeSeries(past=2, use_all_past=True)
        nx, ny, nw = build_ts_X_y(bs, X, y, weights)
        print('X=', X)
        print('y=', y)
        print('nx=', nx)
        print('ny=', ny)
    """
    if not hasattr(model, "use_all_past") or not hasattr(model, "past"):
        raise TypeError(  # pragma: no cover
            "model must be of type BaseTimeSeries not {}".format(type(model)))
    if same_rows:
        if model.use_all_past:
            ncol = X.shape[1] if X is not None else 0
            nrow = y.shape[0] - model.delay2 - model.past + 2

            new_X = numpy.full(
                (y.shape[0], ncol * model.past + model.past), numpy.nan, dtype=y.dtype)
            first = y.shape[0] - nrow
            if X is not None:
                for i in range(0, model.past):
                    begin = i * ncol
                    end = begin + ncol
                    new_X[i:, begin:end] = X[i:]
            for i in range(0, model.past):
                end = y.shape[0] + i + model.delay1 - 1 - model.delay2
                new_X[first - i:first - i + end - i,
                      i + ncol * model.past] = y[i: end]

            new_y = numpy.full(
                (y.shape[0], model.delay2 - model.delay1), numpy.nan, dtype=y.dtype)
            for i in range(model.delay1, model.delay2):
                new_y[first:, i - model.delay1] = y[i + 1:i + nrow + 1]

            new_weights = weights
        else:
            ncol = X.shape[1] if X is not None else 0
            nrow = y.shape[0] - model.delay2 - model.past + 2
            first = y.shape[0] - nrow

            new_X = numpy.full(
                (y.shape[0], ncol + model.past), numpy.nan, dtype=y.dtype)
            if X is not None:
                new_X[first:, :X.shape[1]] = (
                    X[model.past - 1: X.shape[0] - model.delay2 + 1])
            for i in range(model.past):
                end = y.shape[0] + i + model.delay1 - \
                    1 - model.delay2 - model.past + 2
                new_X[first:, i + ncol] = y[i: end]

            new_y = numpy.full(
                (y.shape[0], model.delay2 - model.delay1), numpy.nan, dtype=y.dtype)
            for i in range(model.delay1, model.delay2):
                dec = model.past - 1
                new_y[first:, i - model.delay1] = y[i + dec:i + nrow + dec]
            new_weights = weights
    else:
        if model.use_all_past:
            ncol = X.shape[1] if X is not None else 0
            nrow = y.shape[0] - model.delay2 - model.past + 2

            new_X = numpy.empty(
                (nrow, ncol * model.past + model.past), dtype=y.dtype)
            if X is not None:
                for i in range(0, model.past):
                    begin = i * ncol
                    end = begin + ncol
                    new_X[:, begin:end] = X[i: i + nrow]
            for i in range(0, model.past):
                end = y.shape[0] + i + model.delay1 - 1 - model.delay2
                new_X[:, i + ncol * model.past] = y[i: end]

            new_y = numpy.empty(
                (nrow, model.delay2 - model.delay1), dtype=y.dtype)
            for i in range(model.delay1, model.delay2):
                new_y[:, i - model.delay1] = y[i + 1:i + nrow + 1]

            new_weights = (None if weights is None
                           else weights[model.past - 1:model.past - 1 + nrow])
        else:
            ncol = X.shape[1] if X is not None else 0
            nrow = y.shape[0] - model.delay2 - model.past + 2

            new_X = numpy.empty((nrow, ncol + model.past), dtype=y.dtype)
            if X is not None:
                new_X[:, :X.shape[1]] = X[model.past -
                                          1: X.shape[0] - model.delay2 + 1]
            for i in range(model.past):
                end = y.shape[0] + i + model.delay1 - \
                    1 - model.delay2 - model.past + 2
                new_X[:, i + ncol] = y[i: end]

            new_y = numpy.empty(
                (nrow, model.delay2 - model.delay1), dtype=y.dtype)
            for i in range(model.delay1, model.delay2):
                dec = model.past - 1
                new_y[:, i - model.delay1] = y[i + dec:i + nrow + dec]
            new_weights = (None if weights is None
                           else weights[model.past - 1:model.past - 1 + nrow])
    return new_X, new_y, new_weights


def check_ts_X_y(model, X, y):
    """
    Checks that datasets *(X, y)* was built with function
    @see fn build_ts_X_y.
    """
    cfg = get_config()
    if cfg.get('assume_finite', True):
        return  # pragma: no cover
    if X.dtype not in (numpy.float32, numpy.float64):
        raise TypeError(
            "Features must be of type float32 and float64 not {}.".format(X.dtype))
    if y is not None and y.dtype not in (numpy.float32, numpy.float64):
        raise TypeError(  # pragma: no cover
            "Features must be of type float32 and float64 not {}.".format(y.dtype))
    cst = model.past
    if (hasattr(model, 'preprocessing_') and model.preprocessing_ is not None):
        cst += model.preprocessing_.context_length
    if y is None:
        if cst > 0:
            raise AssertionError(  # pragma: no cover
                "y must be specified to give the model past data to predict, "
                "it requires at least {} observations.".format(cst))
        return  # pragma: no cover
    if y.shape[0] != X.shape[0]:
        raise AssertionError(  # pragma: no cover
            "X and y must have the same number of rows {} != {}.".format(
                X.shape[0], y.shape[0]))
    if len(y.shape) > 1 and y.shape[1] != 1:
        raise AssertionError(  # pragma: no cover
            "y must be 1-dimensional not has shape {}.".format(y.shape))
    if y.shape[0] < cst:
        raise AssertionError(  # pragma: no cover
            "y is not enough past data to predict, "
            "it requires at least {} observations.".format(cst))



"""
@file
@brief Combines a *k-means* followed by a predictor.
"""
import textwrap
import inspect
import numpy
from sklearn.linear_model import LogisticRegression
from sklearn.cluster import KMeans
from sklearn.base import BaseEstimator, ClassifierMixin, clone


class ClassifierAfterKMeans(BaseEstimator, ClassifierMixin):
    """
    Applies a *k-means* (see :epkg:`sklearn:cluster:KMeans`)
    for each class, then adds the distance to each cluster
    as a feature for a classifier.
    See notebook :ref:`logisticregressionclusteringrst`.
    """

    def __init__(self, estimator=None, clus=None, **kwargs):
        """
        @param  estimator   :epkg:`sklearn:linear_model:LogisiticRegression`
                            by default
        @param  clus        clustering applied on each class,
                            by default k-means with two classes
        @param  kwargs      sent to :meth:`set_params
                            <mlinsights.mlmodel.classification_kmeans.
                            ClassifierAfterKMeans.set_params>`,
                            see its documentation to understand how to
                            specify parameters
        """
        ClassifierMixin.__init__(self)
        BaseEstimator.__init__(self)
        if estimator is None:
            estimator = LogisticRegression()
        if clus is None:
            clus = KMeans(n_clusters=2)
        self.estimator = estimator
        self.clus = clus
        if not hasattr(clus, "transform"):
            raise AttributeError(  # pragma: no cover
                "clus does not have a transform method.")
        if kwargs:
            self.set_params(**kwargs)

    def fit(self, X, y, sample_weight=None):
        """
        Runs a *k-means* on each class
        then trains a classifier on the
        extended set of features.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :param y: numpy array of shape [n_samples, n_targets]
            Target values. Will be cast to X's dtype if necessary
        :param sample_weight: numpy array of shape [n_samples]
            Individual weights for each sample
        :return: self : returns an instance of self.

        Fitting attributes:
        * `labels_`: dictionary of clustering models
        * `clus_`: array of clustering models
        * `estimator_`: trained classifier
        """
        classes = set(y)
        self.labels_ = list(sorted(classes))
        self.clus_ = {}
        sig = inspect.signature(self.clus.fit)
        for cl in classes:
            m = clone(self.clus)
            Xcl = X[y == cl]
            if sample_weight is None or 'sample_weight' not in sig.parameters:
                w = None
                m.fit(Xcl)
            else:
                w = sample_weight[y == cl]
                m.fit(Xcl, sample_weight=w)
            self.clus_[cl] = m

        extX = self.transform_features(X)
        self.estimator_ = self.estimator.fit(
            extX, y, sample_weight=sample_weight)
        return self

    def transform_features(self, X):
        """
        Applies all the clustering objects
        on every observations and extends the list of
        features.

        @param      X       features
        @return             extended features
        """
        preds = []
        for _, v in sorted(self.clus_.items()):
            p = v.transform(X)
            preds.append(p)
        return numpy.hstack(preds)

    def predict(self, X):
        """
        Runs the predictions.
        """
        extX = self.transform_features(X)
        return self.estimator.predict(extX)

    def predict_proba(self, X):
        """
        Converts predictions into probabilities.
        """
        extX = self.transform_features(X)
        return self.estimator.predict_proba(extX)

    def decision_function(self, X):
        """
        Calls *decision_function*.
        """
        extX = self.transform_features(X)
        return self.estimator.decision_function(extX)

    def get_params(self, deep=True):
        """
        Returns the parameters for both
        the clustering and the classifier.

        @param      deep        unused here
        @return                 dict

        :meth:`set_params <mlinsights.mlmodel.classification_kmeans.
        ClassifierAfterKMeans.set_params>`
        describes the pattern parameters names follow.
        """
        res = {}
        for k, v in self.clus.get_params().items():
            res["c_" + k] = v
        for k, v in self.estimator.get_params().items():
            res["e_" + k] = v
        return res

    def set_params(self, **values):
        """
        Sets the parameters before training.
        Every parameter prefixed by ``'e_'`` is an estimator
        parameter, every parameter prefixed by ``'c_'`` is for
        the :epkg:`sklearn:cluster:KMeans`.

        @param      values      valeurs
        @return                 dict
        """
        pc, pe = {}, {}
        for k, v in values.items():
            if k.startswith('e_'):
                pe[k[2:]] = v
            elif k.startswith('c_'):
                pc[k[2:]] = v
            else:
                raise ValueError(  # pragma: no cover
                    "Unexpected parameter name '{0}'".format(k))
        self.clus.set_params(**pc)
        self.estimator.set_params(**pe)

    def __repr__(self):  # pylint: disable=W0222
        """
        Overloads `repr` as *scikit-learn* now relies
        on the constructor signature.
        """
        el = ', '.join(['%s=%r' % (k, v)
                        for k, v in self.get_params().items()])
        text = "%s(%s)" % (self.__class__.__name__, el)
        lines = textwrap.wrap(text, subsequent_indent='    ')
        return "\n".join(lines)



"""
@file
@brief Timeseries preprocessing.
"""
import numpy
from .base import BaseReciprocalTimeSeriesTransformer


class TimeSeriesDifference(BaseReciprocalTimeSeriesTransformer):
    """
    Computes timeseries differences.
    """

    def __init__(self, degree=1):
        """
        @param      degree      number of differences
        """
        BaseReciprocalTimeSeriesTransformer.__init__(self, degree)

    @property
    def degree(self):
        """
        Returns the degree.
        """
        return self.context_length

    def fit(self, X, y, sample_weight=None):
        """
        Stores the first values.
        """
        self.X_ = X[:self.degree].copy()
        self.y_ = y[:self.degree].copy()
        for n in range(1, self.degree):
            self.y_[n:] -= self.y_[n - 1:-1]
        return self

    def transform(self, X, y, sample_weight=None):
        """
        Transforms both *X* and *y*.
        Returns *X* and *y*, returns
        *sample_weight* as well if not None.
        """
        for _ in range(self.degree):
            y = y[1:] - y[:-1]
            X = X[1:]
        if sample_weight is None:
            return X, y
        return X, y, sample_weight[1:]

    def get_fct_inv(self):
        """
        Returns the reverse tranform.
        """
        return TimeSeriesDifferenceInv(self).fit()


class TimeSeriesDifferenceInv(BaseReciprocalTimeSeriesTransformer):
    """
    Computes the reverse of @see cl TimeSeriesDifference.
    """

    def __init__(self, estimator):
        """
        @param      estimator   of type @see cl TimeSeriesDifference
        """
        BaseReciprocalTimeSeriesTransformer.__init__(
            self, estimator.context_length)
        if not isinstance(estimator, TimeSeriesDifference):
            raise TypeError(  # pragma: no cover
                "estimator must be of type TimeSeriesDifference not {}"
                "".format(type(estimator)))
        self.estimator = estimator

    def fit(self, X=None, y=None, sample_weight=None):
        """
        Checks that estimator is fitted.
        """
        if not hasattr(self.estimator, 'X_'):
            raise RuntimeError(  # pragma: no cover
                "Estimator is not fitted.")
        self.estimator_ = self.estimator
        return self

    def transform(self, X, y, sample_weight=None):
        """
        Transforms both *X* and *y*.
        Returns *X* and *y*, returns
        *sample_weight* as well if not None.
        """
        if len(y.shape) == 1:
            y = y.reshape((y.shape[0], 1))
            squeeze = True
        else:
            squeeze = False
        if len(self.estimator_.y_.shape) == 1:
            y0 = self.estimator_.y_.reshape((y.shape[0], 1))
        else:
            y0 = self.estimator_.y_
        r0 = self.estimator_.X_.shape[0]

        nx = numpy.empty((r0 + X.shape[0], X.shape[1]), dtype=X.dtype)
        nx[:r0, :] = self.estimator_.X_
        nx[r0:, :] = X

        ny = numpy.empty((r0 + X.shape[0], y.shape[1]), dtype=X.dtype)
        ny[:r0, :] = y0
        ny[r0:, :] = y

        for i in range(self.estimator_.degree):
            numpy.cumsum(ny[r0 - i - 1:, :], axis=0, out=ny[r0 - i - 1:, :])
        if squeeze:
            ny = numpy.squeeze(ny)
        if sample_weight is None:
            return nx, ny
        nw = numpy.zeros(ny.shape[0])
        de = nw.shape[0] - sample_weight.shape[0]
        nw[de:] = sample_weight
        return nx, ny, nw



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a kind of piecewise linear regression by modifying
the criterion used by the algorithm which builds a decision tree.
"""
import numpy
from sklearn.tree import DecisionTreeRegressor


class PiecewiseTreeRegressor(DecisionTreeRegressor):
    """
    Implements a kind of piecewise linear regression by modifying
    the criterion used by the algorithm which builds a decision tree.
    See :epkg:`sklearn:tree:DecisionTreeRegressor` to get the meaning
    of the parameters except criterion:

    * ``mselin``: optimizes for a piecewise linear regression
    * ``simple``: optimizes for a stepwise regression (equivalent to *mse*)
    """

    def __init__(self, criterion='mselin', splitter='best', max_depth=None,
                 min_samples_split=2, min_samples_leaf=1,
                 min_weight_fraction_leaf=0.0, max_features=None,
                 random_state=None, max_leaf_nodes=None,
                 min_impurity_decrease=0.0, min_impurity_split=None):
        DecisionTreeRegressor.__init__(
            self, criterion=criterion,
            splitter=splitter, max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features, random_state=random_state,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            min_impurity_split=min_impurity_split)

    def fit(self, X, y, sample_weight=None, check_input=True,
            X_idx_sorted=None):
        """
        Replaces the string stored in criterion by an instance of a class.
        """
        replace = None
        if isinstance(self.criterion, str):
            if self.criterion == 'mselin':
                from .piecewise_tree_regression_criterion_linear import (  # pylint: disable=E0611,C0415
                    LinearRegressorCriterion)
                replace = self.criterion
                self.criterion = LinearRegressorCriterion(X)
            elif self.criterion == "simple":
                from .piecewise_tree_regression_criterion_fast import (  # pylint: disable=E0611,C0415
                    SimpleRegressorCriterionFast)
                replace = self.criterion
                self.criterion = SimpleRegressorCriterionFast(X)
        else:
            replace = None

        DecisionTreeRegressor.fit(self, X, y, sample_weight=sample_weight, check_input=check_input,
                                  X_idx_sorted=X_idx_sorted)

        if replace:
            self.criterion = replace

        if self.criterion == "mselin":
            self._fit_reglin(X, y, sample_weight)
        return self

    def _mapping_train(self, X):
        tree = self.tree_
        leaves = [i for i in range(len(tree.children_left))
                  if tree.children_left[i] <= i and tree.children_right[i] <= i]  # pylint: disable=E1136
        dec_path = self.decision_path(X)
        association = numpy.zeros((X.shape[0],))
        association[:] = -1
        mapping = {}
        ntree = 0
        for j in leaves:
            ind = dec_path[:, j] == 1
            ind = numpy.asarray(ind.todense()).flatten()
            if not numpy.any(ind):
                # No training example for this bucket.
                continue
            mapping[j] = ntree
            association[ind] = ntree
            ntree += 1
        return mapping

    def predict_leaves(self, X):
        """
        Returns the leave index for each observation of *X*.

        :param X: array
        :return: array
            leaves index in ``self.leaves_index_``
        """
        # The creation of the sparse matrix could be avoided.
        leaves = self.decision_path(X)
        leaves = leaves[:, self.leaves_index_]
        mat = numpy.argmax(leaves, 1)
        res = numpy.asarray(mat).ravel()
        return res

    def _fit_reglin(self, X, y, sample_weight):
        """
        Fits linear regressions for all leaves.
        Sets attributes ``leaves_mapping_``, ``betas_``, ``leaves_index_``.
        The first attribute is a dictionary ``{leave: row}``
        which maps a leave of the tree to the coefficients
        ``betas_[row, :]`` of a regression trained on all training
        points mapped a specific leave. ``leaves_index_`` keeps
        in memory a set of leaves.
        """
        from .piecewise_tree_regression_criterion_linear import (  # pylint: disable=E0611,C0415
            LinearRegressorCriterion)

        tree = self.tree_
        self.leaves_index_ = [i for i in range(len(tree.children_left))
                              if tree.children_left[i] <= i and tree.children_right[i] <= i]  # pylint: disable=E1136
        if tree.n_leaves != len(self.leaves_index_):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of leaves {} != {}".format(
                    tree.n_leaves, len(self.leaves_index_)))
        pred_leaves = self.predict_leaves(X)
        self.leaves_mapping_ = {k: i for i, k in enumerate(pred_leaves)}
        self.betas_ = numpy.empty((len(self.leaves_index_), X.shape[1] + 1))
        for i, _ in enumerate(self.leaves_index_):
            ind = pred_leaves == i
            xs = X[ind, :].copy()
            ys = y[ind].astype(numpy.float64)
            if len(ys.shape) == 1:
                ys = ys[:, numpy.newaxis]
            ys = ys.copy()
            ws = sample_weight[ind].copy() if sample_weight else None
            dec = LinearRegressorCriterion.create(xs, ys, ws)
            dec.node_beta(self.betas_[i, :])

    def predict(self, X, check_input=True):
        """
        Overloads method *predict*. Falls back into
        the predict from a decision tree is criterion is
        *mse*, *mae*, *simple*. Computes the predictions
        from linear regression if the criterion is *mselin*.
        """
        if self.criterion == 'mselin':
            return self._predict_reglin(X, check_input=check_input)
        return DecisionTreeRegressor.predict(self, X, check_input=check_input)

    def _predict_reglin(self, X, check_input=True):
        """
        Computes the predictions with a linear regression
        fitted with the observations mapped to each leave
        of the tree.

        :param X: array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.
        :param check_input: boolean, (default=True)
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.
        :return: y, array of shape = [n_samples] or [n_samples, n_outputs]
            The predicted classes, or the predict values.
        """
        leaves = self.predict_leaves(X)
        pred = numpy.ones((X.shape[0], 1))
        Xone = numpy.hstack([X, pred])
        for i in range(0, X.shape[0]):
            li = leaves[i]
            pred[i] = numpy.dot(Xone[i, :], self.betas_[li, :])
        return pred.ravel()



# See _piecewise_tree_regression_common.pyx for implementation details.
cimport cython
import numpy
cimport numpy

from sklearn.tree._criterion cimport Criterion
from sklearn.tree._criterion cimport SIZE_t, DOUBLE_t


cdef class CommonRegressorCriterion(Criterion):

    cdef const DOUBLE_t[:, ::1] sample_X

    cdef void _update_weights(self, SIZE_t start, SIZE_t end,
                              SIZE_t old_pos, SIZE_t new_pos) nogil

    cdef void _mean(self, SIZE_t start, SIZE_t end, DOUBLE_t *mean,
                    DOUBLE_t *weight) nogil
    cdef double _mse(self, SIZE_t start, SIZE_t end, DOUBLE_t mean,
                     DOUBLE_t weight) nogil

    cdef void children_impurity_weights(self, double* impurity_left,
                                        double* impurity_right,
                                        double* weight_left,
                                        double* weight_right) nogil



"""
@file
@brief Implements a transform which modifies the target
and applies the reverse transformation on the target.
"""
import numpy
from sklearn.exceptions import NotFittedError
from sklearn.neighbors import NearestNeighbors
from .sklearn_transform_inv import BaseReciprocalTransformer


class FunctionReciprocalTransformer(BaseReciprocalTransformer):
    """
    The transform is used to apply a function on a the target,
    predict, then transform the target back before scoring.
    The transforms implements a series of predefined functions:

    .. runpython::
        :showcode:

        import pprint
        from mlinsights.mlmodel.sklearn_transform_inv_fct import FunctionReciprocalTransformer
        pprint.pprint(FunctionReciprocalTransformer.available_fcts())
    """

    @staticmethod
    def available_fcts():
        """
        Returns the list of predefined functions.
        """
        return {
            'log': (numpy.log, 'exp'),
            'exp': (numpy.exp, 'log'),
            'log(1+x)': (lambda x: numpy.log(x + 1), 'exp(x)-1'),
            'log1p': (numpy.log1p, 'expm1'),
            'exp(x)-1': (lambda x: numpy.exp(x) - 1, 'log'),
            'expm1': (numpy.expm1, 'log1p'),
        }

    def __init__(self, fct, fct_inv=None):
        """
        @param      fct         function name of numerical function
        @param      fct_inv     optional if *fct* is a function name,
                                reciprocal function otherwise
        """
        BaseReciprocalTransformer.__init__(self)
        if isinstance(fct, str):
            if fct_inv is not None:
                raise ValueError(
                    "If fct is a function name, fct_inv must not be specified.")
            opts = self.__class__.available_fcts()
            if fct not in opts:
                raise ValueError("Unknown fct '{}', it should in {}.".format(
                    fct, list(sorted(opts))))
        else:
            if fct_inv is None:
                raise ValueError(
                    "If fct is callable, fct_inv must be specified.")
        self.fct = fct
        self.fct_inv = fct_inv

    def fit(self, X=None, y=None, sample_weight=None):
        """
        Just defines *fct* and *fct_inv*.
        """
        if callable(self.fct):
            self.fct_ = self.fct
            self.fct_inv_ = self.fct_inv
        else:
            opts = self.__class__.available_fcts()
            self.fct_, self.fct_inv_ = opts[self.fct]
        return self

    def get_fct_inv(self):
        """
        Returns a trained transform which reverse the target
        after a predictor.
        """
        if isinstance(self.fct_inv_, str):
            res = FunctionReciprocalTransformer(self.fct_inv_)
        else:
            res = FunctionReciprocalTransformer(self.fct_inv_, self.fct_)
        return res.fit()

    def transform(self, X, y):
        """
        Transforms *X* and *y*.
        Returns transformed *X* and *y*.
        If *y* is None, the returned value for *y*
        is None as well.
        """
        if y is None:
            return X, None
        return X, self.fct_(y)


class PermutationReciprocalTransformer(BaseReciprocalTransformer):
    """
    The transform is used to permute targets,
    predict, then permute the target back before scoring.
    nan values remain nan values. Once fitted, the transform
    has attribute ``permutation_`` which keeps
    track of the permutation to apply.
    """

    def __init__(self, random_state=None, closest=False):
        """
        @param      random_state    random state
        @param      closest         if True, finds the closest permuted element
        """
        BaseReciprocalTransformer.__init__(self)
        self.random_state = random_state
        self.closest = closest

    def fit(self, X=None, y=None, sample_weight=None):
        """
        Defines a random permutation over the targets.
        """
        if y is None:
            raise RuntimeError("targets cannot be empty.")
        num = numpy.issubdtype(y.dtype, numpy.floating)
        perm = {}
        for u in y.ravel():
            if num and numpy.isnan(u):
                continue
            if u in perm:
                continue
            perm[u] = len(perm)

        lin = numpy.arange(len(perm))
        if self.random_state is None:
            lin = numpy.random.permutation(lin)
        else:
            rs = numpy.random.RandomState(  # pylint: disable=E1101
                self.random_state)  # pylint: disable=E1101
            lin = rs.permutation(lin)

        for u in perm:
            perm[u] = lin[perm[u]]
        self.permutation_ = perm

    def _check_is_fitted(self):
        if not hasattr(self, 'permutation_'):
            raise NotFittedError(
                "This instance {} is not fitted yet. Call 'fit' with "
                "appropriate arguments before using this method.".format(
                    type(self)))

    def get_fct_inv(self):
        """
        Returns a trained transform which reverse the target
        after a predictor.
        """
        self._check_is_fitted()
        res = PermutationReciprocalTransformer(
            self.random_state, closest=self.closest)
        res.permutation_ = {v: k for k, v in self.permutation_.items()}
        return res

    def _find_closest(self, cl):
        if not hasattr(self, 'knn_'):
            self.knn_ = NearestNeighbors(n_neighbors=1, algorithm='kd_tree')
            self.knn_perm_ = numpy.array(list(self.permutation_))
            self.knn_perm_ = self.knn_perm_.reshape((len(self.knn_perm_), 1))
            self.knn_.fit(self.knn_perm_)
        ind = self.knn_.kneighbors([[cl]], return_distance=False)
        res = self.knn_perm_[ind, 0]
        if self.knn_perm_.dtype in (numpy.float32, numpy.float64):
            return float(res)
        if self.knn_perm_.dtype in (numpy.int32, numpy.int64):
            return int(res)
        raise NotImplementedError("The function does not work for type {}.".format(
            self.knn_perm_.dtype))

    def transform(self, X, y):
        """
        Transforms *X* and *y*.
        Returns transformed *X* and *y*.
        If *y* is None, the returned value for *y*
        is None as well.
        """
        if y is None:
            return X, None
        self._check_is_fitted()
        if len(y.shape) == 1 or y.dtype in (numpy.str, numpy.int32, numpy.int64):
            # permutes classes
            yp = y.copy().ravel()
            num = numpy.issubdtype(y.dtype, numpy.floating)
            for i in range(len(yp)):  # pylint: disable=C0200
                if num and numpy.isnan(yp[i]):
                    continue
                if yp[i] not in self.permutation_:
                    if self.closest:
                        cl = self._find_closest(yp[i])
                    else:
                        raise RuntimeError("Unable to find key '{}' in {}.".format(
                            yp[i], list(sorted(self.permutation_))))
                else:
                    cl = yp[i]
                yp[i] = self.permutation_[cl]
            return X, yp.reshape(y.shape)
        else:
            # y is probababilies or raw score
            if len(y.shape) != 2:
                raise RuntimeError(
                    "yp should be a matrix but has shape {}.".format(y.shape))
            cl = [(v, k) for k, v in self.permutation_.items()]
            cl.sort()
            new_perm = {}
            for cl, current in cl:
                new_perm[current] = len(new_perm)
            yp = y.copy()
            for i in range(y.shape[1]):
                yp[:, new_perm[i]] = y[:, i]
            return X, yp



"""
@file
@brief Timeseries metrics.
"""
import numpy


def ts_mape(expected_y, predicted_y, sample_weight=None):
    """
    Computes :math:`\\frac{\\sum_i | \\hat{Y_t} - Y_t |}
    {\\sum_i | Y_t - Y_{t-1} |}`.
    It compares the prediction to what a dummy
    predictor would do by using the previous day
    as a prediction.

    @param      expected_y          expected values
    @param      predicted_y         predictions
    @return                         metrics
    """
    if len(expected_y) != len(predicted_y):
        raise ValueError(  # pragma: no cover
            'Size mismatch {} != {}.'.format(
                len(expected_y), len(predicted_y)))
    expected_y = numpy.squeeze(expected_y)
    predicted_y = numpy.squeeze(predicted_y)
    mask = numpy.isnan(predicted_y)
    mask2 = mask.copy()
    mask2[1:] |= numpy.isnan(predicted_y[:-1])
    expected_y = numpy.ma.masked_array(expected_y, mask=mask)
    predicted_y = numpy.ma.masked_array(predicted_y, mask=mask2)
    if sample_weight is None:
        dy1 = numpy.sum(numpy.abs(expected_y[:-1] - expected_y[1:]))
        dy2 = numpy.sum(numpy.abs(predicted_y[1:] - expected_y[1:]))
    else:
        dy1 = numpy.sum(
            (numpy.abs(expected_y[:-1] - expected_y[1:]) * sample_weight[1:]))
        dy2 = numpy.sum(
            (numpy.abs(predicted_y[1:] - expected_y[1:]) * sample_weight[1:]))
    dy1 = dy1.sum()
    dy2 = dy2.sum()
    if dy1 == 0:
        return 0 if dy2 == 0 else numpy.infty
    return dy2 / dy1



"""
@file
@brief Data aggregation for timeseries.
"""
import datetime
import pandas
from pandas.tseries.frequencies import to_offset


def _get_column_name(df, name='agg'):
    """
    Returns a unique column name not in the existing dataframe.

    @param      df      dataframe
    @param      name    prefix
    @return             new column name
    """
    while name in df.columns:
        name += '_'
    return name


def aggregate_timeseries(df, index='time', values='y',
                         unit='half-hour', agg='sum',
                         per=None):
    """
    Aggregates timeseries assuming the data is in a dataframe.

    @param      df      dataframe
    @param      index   time column
    @param      values  value or values column
    @param      unit    aggregate over a specific period
    @param      sum     kind of aggregation
    @param      per     second aggregation, per week...
    @return             aggregated values
    """
    if df is None:
        if len(values.shape) == 1:
            df = pandas.DataFrame(dict(time=index, y=values))
            values = 'y'
        else:
            df = pandas.DataFrame(dict(time=index))
            for i in range(values.shape[1]):
                df['y%d' % i] = values[:, i]
            values = list(df.columns)[1:]
        index = 'time'

    def round_(serie, freq, per):
        fr = to_offset(freq)
        res = pandas.DatetimeIndex(serie).floor(fr)  # pylint: disable=E1101
        if per is None:
            return res
        if per == 'week':
            pyres = res.to_pydatetime()
            return pandas.to_timedelta(
                map(
                    lambda t: datetime.timedelta(
                        days=t.weekday(), hours=t.hour, minutes=t.minute),
                    pyres))
        if per == 'month':
            pyres = res.to_pydatetime()
            return pandas.to_timedelta(
                map(
                    lambda t: datetime.timedelta(
                        days=t.day, hours=t.hour, minutes=t.minute),
                    pyres))
        raise ValueError(  # pragma: no cover
            "Unknown frequency '{}'.".format(per))

    agg_name = _get_column_name(df)
    df = df.copy()
    if unit == 'half-hour':
        freq = datetime.timedelta(minutes=30)
        df[agg_name] = round_(df[index], freq, per)
    else:
        raise ValueError(  # pragma: no cover
            "Unknown time unit '{}'.".format(unit))
    if not isinstance(values, list):
        values = [values]
    if agg == 'sum':
        gr = df[[agg_name] + values].groupby(agg_name, as_index=False).sum()
        agg_name = _get_column_name(gr, 'week' + index)
        gr.columns = [agg_name] + list(gr.columns[1:])
    elif agg == 'norm':
        gr = df[[agg_name] + values].groupby(agg_name, as_index=False).sum()
        agg_name = _get_column_name(gr, 'week' + index)
        agg_cols = list(gr.columns[1:])
        gr.columns = [agg_name] + agg_cols
        for c in agg_cols:
            su = gr[c].sum()
            if su != 0:
                gr[c] /= su
    else:
        raise ValueError(  # pragma: no cover
            "Unknown aggregation '{}'.".format(agg))
    return gr.sort_values(agg_name).reset_index(drop=True)



"""
@file
@brief Implements new features such as polynomial features.
"""
import numpy
from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from ._extended_features_polynomial import _transform_iall, _transform_ionly, _combinations_poly


class ExtendedFeatures(BaseEstimator, TransformerMixin):
    """
    Generates extended features such as polynomial features.

    :param kind: string
        ``'poly'`` for polynomial features,
        ``'poly-slow'`` for polynomial features in *scikit-learn 0.20.2*
    :param poly_degree: integer
        The degree of the polynomial features. Default = 2.
    :param poly_interaction_only: boolean
        If true, only interaction features are produced: features that
        are products of at most degree distinct input features
        (so not ``x[1] ** 2, x[0] * x[2] ** 3``, etc.).
    :param poly_include_bias: boolean
        If True (default), then include a bias column, the feature in
        which all polynomial powers are zero (i.e. a column of ones -
        acts as an intercept term in a linear model).

    Fitted attributes:

    * `n_input_features_`: int
        The total number of input features.
    * `n_output_features_`: int
        The total number of polynomial output features. The number of output
        features is computed by iterating over all suitably sized combinations
        of input features.
    """

    def __init__(self, kind='poly', poly_degree=2, poly_interaction_only=False,
                 poly_include_bias=True):
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)
        self.kind = kind
        self.poly_degree = poly_degree
        self.poly_include_bias = poly_include_bias
        self.poly_interaction_only = poly_interaction_only

    def get_feature_names(self, input_features=None):
        """
        Returns feature names for output features.

        :param input_features: list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.
        :return: output_feature_names : list of string, length n_output_features
        """
        if self.kind == 'poly':
            return self._get_feature_names_poly(input_features)
        if self.kind == 'poly-slow':
            return self._get_feature_names_poly(input_features)
        raise ValueError(  # pragma: no cover
            "Unknown extended features '{}'.".format(self.kind))

    def _get_feature_names_poly(self, input_features=None):
        """
        Returns feature names for output features for
        the polynomial features.
        """
        if input_features is None:
            input_features = ["x%d" %
                              i for i in range(0, self.n_input_features_)]
        elif len(input_features) != self.n_input_features_:
            raise ValueError(
                "input_features should contain {} strings.".format(
                    self.n_input_features_))

        names = ["1"] if self.poly_include_bias else []
        n = self.n_input_features_
        interaction_only = self.poly_interaction_only
        for d in range(0, self.poly_degree):
            if d == 0:
                pos = len(names)
                names.extend(input_features)
                index = list(range(pos, len(names)))
                index.append(len(names))
            else:
                new_index = []
                end = index[-1]
                for i in range(0, n):
                    a = index[i]
                    new_index.append(len(names))
                    start = a + (index[i + 1] - index[i]
                                 if interaction_only else 0)
                    names.extend([a + " " + input_features[i]
                                  for a in names[start:end]])
                new_index.append(len(names))
                index = new_index

        def process_name(col):
            scol = col.split()
            res = []
            for c in sorted(scol):
                if len(res) == 0 or res[-1][0] != c:
                    res.append((c, 1))
                else:
                    res[-1] = (c, res[-1][1] + 1)
            return " ".join(["%s^%d" % r if r[1] > 1 else r[0] for r in res])

        names = [process_name(s) for s in names]
        return names

    def fit(self, X, y=None):
        """
        Compute number of output features.

        :param X: array-like, shape (n_samples, n_features)
            The data.
        :return: self : instance
        """
        self.n_input_features_ = X.shape[1]
        self.n_output_features_ = len(self.get_feature_names())

        if self.kind == 'poly':
            return self._fit_poly(X, y)
        elif self.kind == 'poly-slow':
            return self._fit_poly(X, y)
        raise ValueError(
            "Unknown extended features '{}'.".format(self.kind))

    def _fit_poly(self, X, y=None):
        """
        Fitting method for the polynomial features.
        """
        check_array(X, accept_sparse=False)
        return self

    def transform(self, X):
        """
        Transforms data to extended features.

        :param X: array-like, shape [n_samples, n_features]
            The data to transform, row by row.
            rns
        :param XP: numpy.ndarray, shape [n_samples, NP]
            The matrix of features, where NP is the number of polynomial
            features generated from the combination of inputs.
        """
        n_features = X.shape[1]
        if n_features != self.n_input_features_:
            raise ValueError("X shape does not match training shape")
        if self.kind == 'poly':
            return self._transform_poly(X)
        if self.kind == 'poly-slow':
            return self._transform_poly_slow(X)
        raise ValueError(  # pragma: no cover
            "Unknown extended features '{}'.".format(self.kind))

    def _transform_poly(self, X):
        """
        Transforms data to polynomial features.
        """
        if sparse.isspmatrix(X):
            raise NotImplementedError(  # pragma: no cover
                "Not implemented for sparse matrices.")

        XP = numpy.empty(
            (X.shape[0], self.n_output_features_), dtype=X.dtype)

        def multiply(A, B, C):
            return numpy.multiply(A, B, out=C)

        def final(X):
            return X

        if self.poly_interaction_only:
            return _transform_ionly(self.poly_degree, self.poly_include_bias,
                                    XP, X, multiply, final)
        return _transform_iall(self.poly_degree, self.poly_include_bias,
                               XP, X, multiply, final)

    def _transform_poly_slow(self, X):
        """
        Transforms data to polynomial features.
        """
        if sparse.isspmatrix(X):
            raise NotImplementedError(  # pragma: no cover
                "Not implemented for sparse matrices.")

        comb = _combinations_poly(X.shape[1], self.poly_degree, self.poly_interaction_only,
                                  include_bias=self.poly_include_bias)
        order = 'C'  # how to get order from X.
        XP = numpy.empty((X.shape[0], self.n_output_features_),
                         dtype=X.dtype, order=order)
        for i, comb in enumerate(comb):
            XP[:, i] = X[:, comb].prod(1)
        return XP



"""
@file
@brief Featurizers for machine learned models.
"""
import sys
import io
import os
import urllib.request
from PIL import Image


def plot_gallery_images(imgs, texts=None, width=4, return_figure=False,
                        ax=None, folder_image=None, **figure):
    """
    Plots a gallery of images using :epkg:`matplotlib`.

    @param      imgs            list of images (filename, urls or :epkg:`Pillow` objects),
    @param      texts           text to display (if None, print ``'img % i'``)
    @param      width           number of images on the same line (unused if *imgs* is a matrix)
    @param      figure          additional parameters when the figure is created
    @param      return_figure   return the figure as well as the axes
    @param      ax              None or existing axes, it should have the same
                                shape of *imgs*
    @param      folder_image    image paths may be relative to some folder,
                                in that case, they should be relative to
                                this folder
    @return                     axes or (figure, axes) if *return_figure* is True

    .. image:: gal.jpg

    See also notebook :ref:`searchimageskerasrst` or
    :ref:`searchimagestorchrst` for an example.
    *imgs* can also be a matrix to force the function to
    use the same coordinates.
    """
    if "plt" not in sys.modules:
        import matplotlib.pyplot as plt  # pylint: disable=C0415

    if hasattr(imgs, 'shape') and len(imgs.shape) == 2:
        height, width = imgs.shape
        if ax is not None and ax.shape != imgs.shape:
            raise ValueError(  # pragma: no cover
                "ax.shape {0} != imgs.shape {1}".format(ax.shape, imgs.shape))
        imgs = imgs.ravel()
        if texts is not None:
            texts = texts.ravel()
    else:
        height = len(imgs) // width
        if len(imgs) % width != 0:
            height += 1

    if ax is None:
        if 'figsize' not in figure:
            figure['figsize'] = (12, height * 3)
        fig, ax = plt.subplots(height, width, **figure)
    elif return_figure:
        raise ValueError(  # pragma: no cover
            "ax is specified and return_figure is True")

    for i, img in enumerate(imgs):
        if img is None:
            continue
        y, x = i // width, i % width
        if height == 1:
            ind = x
        elif width == 1:
            ind = y
        else:
            ind = y, x

        if isinstance(img, str):
            if "//" in img:
                # url
                with urllib.request.urlopen(img) as response:
                    content = response.read()
                try:
                    im = Image.open(io.BytesIO(content))
                except OSError as e:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to read image '{}'.".format(img)) from e
            else:
                # local file
                if folder_image is not None:
                    im = Image.open(os.path.join(folder_image, img))
                else:
                    im = Image.open(img)
        else:
            im = img
        if hasattr(im, 'size'):
            ax[ind].imshow(im)
            if texts is None:
                t = "img %d" % i
            else:
                t = texts[i]
            ax[ind].text(0, 0, t)
        ax[ind].axis('off')

    for i in range(len(imgs), width * height):
        y, x = i // width, i % width
        if height == 1:
            ind = x
        else:
            ind = y, x
        ax[ind].axis('off')

    if return_figure:
        return fig, ax
    return ax



"""
@file
@brief Shortcuts to *mlmodel*.
"""
from .anmf_predictor import ApproximateNMFPredictor
from .categories_to_integers import CategoriesToIntegers
from .classification_kmeans import ClassifierAfterKMeans
from .decision_tree_logreg import DecisionTreeLogisticRegression
from .extended_features import ExtendedFeatures
from .interval_regressor import IntervalRegressor
from .kmeans_constraint import ConstraintKMeans
from .kmeans_l1 import KMeansL1L2
from .ml_featurizer import model_featurizer
from .piecewise_estimator import PiecewiseRegressor, PiecewiseClassifier
from .piecewise_tree_regression import PiecewiseTreeRegressor
from .predictable_tsne import PredictableTSNE
from .quantile_mlpregressor import QuantileMLPRegressor
from .quantile_regression import QuantileLinearRegression
from .sklearn_testing import test_sklearn_pickle, test_sklearn_clone, test_sklearn_grid_search_cv
from .sklearn_text import TraceableTfidfVectorizer, TraceableCountVectorizer
from .sklearn_transform_inv_fct import FunctionReciprocalTransformer, PermutationReciprocalTransformer
from .target_predictors import TransformedTargetClassifier2, TransformedTargetRegressor2
from .transfer_transformer import TransferTransformer
